@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Network.XmlRpc.Internals
@@ -26,15 +27,19 @@ Value(..), Type(..), XmlRpcType(..),
 parseResponse, parseCall, getField, getFieldMaybe,
 -- * Converting to XML
 renderCall, renderResponse,
+-- * Converting to and from DTD types
+toXRValue, fromXRValue,
+toXRMethodCall, fromXRMethodCall,
+toXRMethodResponse, fromXRMethodResponse,
+toXRParams, fromXRParams,
+toXRMember, fromXRMember,
 -- * Error monad
 Err, maybeToM, handleError, ioErrorToErr
 ) where
 
 import           Control.Exception
 import           Control.Monad
-import           Control.Monad.Error
-import           Control.Monad.Identity
-import           Data.Char (chr, ord)
+import           Control.Monad.Except
 import           Data.Char
 import           Data.List
 import           Data.Maybe
@@ -43,13 +48,15 @@ import           Data.Time.Calendar.OrdinalDate (toOrdinalDate)
 import           Data.Time.Calendar.WeekDate (toWeekDate)
 import           Data.Time.Format
 import           Data.Time.LocalTime
-import           Data.Word (Word8)
 import           Numeric (showFFloat)
 import           Prelude hiding (showString, catch)
 import           System.IO.Unsafe (unsafePerformIO)
-import           System.Locale
 import           System.Time (CalendarTime(..))
 
+#if ! MIN_VERSION_time(1,5,0)
+import           System.Locale (defaultTimeLocale)
+#endif
+
 import qualified Data.ByteString.Char8 as BS (ByteString, pack, unpack)
 import qualified Data.ByteString.Lazy.Char8 as BSL (ByteString, pack)
 import qualified Network.XmlRpc.Base64 as Base64
@@ -91,13 +98,14 @@ eitherToM err (Left s)  = fail (err ++ ": " ++ s)
 eitherToM   _ (Right x) = return x
 
 -- | The format for \"dateTime.iso8601\"
+xmlRpcDateFormat :: String
 xmlRpcDateFormat = "%Y%m%dT%H:%M:%S"
 
 --
 -- Error monad stuff
 --
 
-type Err m a = ErrorT String m a
+type Err m a = ExceptT String m a
 
 -- | Evaluate the argument and catch error call exceptions
 errorToErr :: (Show e, MonadError e m) => a -> Err m a
@@ -112,7 +120,7 @@ ioErrorToErr x = (liftIO x >>= return) `catchError` \e -> throwError (show e)
 -- | Handle errors from the error monad.
 handleError :: Monad m => (String -> m a) -> Err m a -> m a
 handleError h m = do
-		  Right x <- runErrorT (catchError m (lift . h))
+		  Right x <- runExceptT (catchError m (lift . h))
 		  return x
 
 errorRead :: (Monad m, Read a) =>
@@ -124,17 +132,6 @@ errorRead r err s = case [x | (x,t) <- r s, ("","") <- lex t] of
 		          [x] -> return x
 		          _   -> fail (err ++ ": '" ++ s ++ "'")
 
--- | Convert an 'Int' to some instance of 'Enum', and fail if the
---   'Int' is out of range.
-errorToEnum :: (Monad m, Bounded a, Enum a) =>
-	       String -- ^ Error message
-	    -> Int
-	    -> Err m a
-errorToEnum err x | x < fromEnum (minBound `asTypeOf` r) = fail err
-		  | x > fromEnum (maxBound `asTypeOf` r) = fail err
-		  | otherwise = return r
-    where r = toEnum x
-
 --
 -- Types for methods calls and responses
 --
@@ -151,9 +148,10 @@ data MethodResponse = Return Value -- ^ A method response returning a value
 
 -- | An XML-RPC value.
 data Value =
-      ValueInt Int -- ^ int or i4
+      ValueInt Int -- ^ int, i4, or i8
     | ValueBool Bool -- ^ bool
     | ValueString String -- ^ string
+    | ValueUnwrapped String -- ^ no inner element
     | ValueDouble Double -- ^ double
     | ValueDateTime LocalTime -- ^ dateTime.iso8601
     | ValueBase64 BS.ByteString -- ^ base 64.  NOTE that you should provide the raw data; the haxr library takes care of doing the base-64 encoding.
@@ -273,6 +271,7 @@ instance XmlRpcType String where
     toValue = ValueString
     fromValue = simpleFromValue f
 	where f (ValueString x) = Just x
+              f (ValueUnwrapped x) = Just x
 	      f _ = Nothing
     getType _ = TString
 
@@ -315,7 +314,7 @@ instance XmlRpcType a => XmlRpcType [(String,a)] where
     toValue xs = ValueStruct [(n, toValue v) | (n,v) <- xs]
 
     fromValue v = case v of
-		  ValueStruct xs -> mapM (\ (n,v) -> liftM ((,) n) (fromValue v)) xs
+		  ValueStruct xs -> mapM (\ (n,v') -> liftM ((,) n) (fromValue v')) xs
 		  _ -> typeError v
     getType _ = TStruct
 
@@ -377,6 +376,7 @@ toXRValue :: Value -> XR.Value
 toXRValue (ValueInt x) = XR.Value [XR.Value_AInt (XR.AInt (showInt x))]
 toXRValue (ValueBool b) = XR.Value [XR.Value_Boolean (XR.Boolean (showBool b))]
 toXRValue (ValueString s) = XR.Value [XR.Value_AString (XR.AString (showString s))]
+toXRValue (ValueUnwrapped s) = XR.Value [XR.Value_Str s]
 toXRValue (ValueDouble d) = XR.Value [XR.Value_ADouble (XR.ADouble (showDouble d))]
 toXRValue (ValueDateTime t) =
    XR.Value [ XR.Value_DateTime_iso8601 (XR.DateTime_iso8601 (showDateTime t))]
@@ -428,7 +428,7 @@ toXRMember (n, v) = XR.Member (XR.Name n) (toXRValue v)
 fromXRValue :: Monad m => XR.Value -> Err m Value
 fromXRValue (XR.Value vs)
   =  case (filter notstr vs) of
-       []     -> liftM  (ValueString . concat) (mapM (readString . unstr) vs)
+       []     -> liftM  (ValueUnwrapped . concat) (mapM (readString . unstr) vs)
        (v:_)  -> f v
   where
   notstr (XR.Value_Str _)  = False
@@ -437,6 +437,7 @@ fromXRValue (XR.Value vs)
   unstr (XR.Value_Str x)  = x
 
   f (XR.Value_I4 (XR.I4 x)) = liftM ValueInt (readInt x)
+  f (XR.Value_I8 (XR.I8 x)) = liftM ValueInt (readInt x)
   f (XR.Value_AInt (XR.AInt x)) = liftM ValueInt (readInt x)
   f (XR.Value_Boolean (XR.Boolean x)) = liftM ValueBool (readBool x)
   f (XR.Value_ADouble (XR.ADouble x)) = liftM ValueDouble (readDouble x)
